import codecs, functools, json, mimetypes, pathlib, io, asyncio, logging, os, typing
from typing import Any, Callable
from collections.abc import Awaitable, Iterable, MutableMapping, AsyncGenerator
from email.utils import formatdate
from pydantic import ValidationError
from rxxxt.helpers import match_path

BytesLike = bytes | bytearray
ASGIHeaders = Iterable[tuple[BytesLike, BytesLike]]

ASGIScope = MutableMapping[str, Any]
ASGIFnSend = Callable[[MutableMapping[str, Any]], Awaitable[Any]]
ASGIFnReceive = Callable[[], Awaitable[MutableMapping[str, Any]]]
ASGIHandler = Callable[[ASGIScope, ASGIFnReceive, ASGIFnSend], Awaitable[Any]]

class ASGINextException(Exception): pass

class TransportContext:
  def __init__(self, scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> None:
    self.scope = scope
    self.receive = receive
    self.send = send

  @property
  def path(self): return self.scope["path"]

  @property
  def query_string(self) -> str | None: return None if not self.scope["query_string"] else self.scope["query_string"].decode("utf-8")

  @property
  def fullpath(self): return (self.scope["raw_path"] or b"").decode("utf-8").split("?", 1)[0]

  @functools.cached_property
  def headers(self):
    res: dict[str, tuple[str, ...]] = {}
    for k, v in self.scope["headers"]:
      key = k.decode(errors="ignore").lower()
      res[key] = res.get(key, ()) + (v.decode(errors="ignore"),)
    return res

  @functools.cached_property
  def content_type(self):
    ct = self.headers.get("content-type")
    if ct is None or len(ct) == 0: raise ValueError("No content type specified on request!")
    if len(ct) > 1: raise ValueError("More than one content-type was specified!")
    ct = ct[0]
    parts = [ p.strip() for p in ct.split(";") ]
    mime_type = parts[0].lower()
    params = { k.lower(): v for k, v in (tuple(p.split("=") for p in parts[1:] if p.count("=") == 1)) }
    return mime_type, params

  @property
  def location(self):
    location = self.path
    if self.query_string is not None: location += f"?{self.query_string}"
    return location

  def next(self): raise ASGINextException()

class WebsocketContext(TransportContext):
  def __init__(self, scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> None:
    super().__init__(scope, receive, send)
    self._connected = True

  @property
  def connected(self): return self._connected

  async def setup(self, headers: ASGIHeaders = (), subprotocol: str | None = None):
    event = await self.receive()
    if event["type"] != "websocket.connect": raise ConnectionError("Did not receive connect event!")
    await self.send({ "type": "websocket.accept", "subprotocol": subprotocol, "headers": [ (name.lower(), value) for name, value in headers ] })

  async def receive_message(self) -> BytesLike | str:
    while self._connected:
      event = await self.receive()
      if event["type"] == "websocket.disconnect":
        self._connected = False
        raise ConnectionError("Connection closed!")
      elif event["type"] == "websocket.receive":
        return event.get("bytes", event.get("text"))
    raise ConnectionError("Connection closed!")

  async def send_message(self, data: str | BytesLike):
    if not self._connected: raise ConnectionError("Not connected!")
    event: dict[str, Any] = { "type": "websocket.send", "bytes": None, "text": None }
    if isinstance(data, str): event["text"] = data
    else: event["bytes"] = data
    await self.send(event)

  async def close(self, code: int = 1000, reason: str = "Normal Closure"):
    await self.send({ "type": "websocket.close", "code": code, "reason": reason })
    self._connected = False

def content_headers(content_length: int, mime_type: str, charset: str | None = None):
  content_type = mime_type
  if charset is not None: content_type += f"; charset={charset}"
  return [
    (b"content-length", str(content_length).encode("utf-8")),
    (b"content-type", content_type.encode("utf-8"))
  ]

class HTTPContext(TransportContext):
  def __init__(self, scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> None:
    super().__init__(scope, receive, send)
    self._response_headers: list[tuple[BytesLike, BytesLike]] = []

  @property
  def method(self): return self.scope["method"]

  def add_response_headers(self, headers: ASGIHeaders): self._response_headers.extend(headers)

  async def response_start(self, status: int, trailers: bool = False):
    await self.send({
      "type": "http.response.start",
      "status": status,
      "headers": self._response_headers,
      "trailers": trailers
    })

  async def response_body(self, data: BytesLike, more_body: bool):
    await self.send({
      "type": "http.response.body",
      "body": data,
      "more_body": more_body
    })

  async def respond_text(self, text: str, status: int = 200, mime_type: str = "text/plain"):
    data = text.encode("utf-8")
    self.add_response_headers(content_headers(len(data), mime_type, "utf-8"))
    await self.response_start(status)
    await self.response_body(data, False)

  async def respond_file(self, path: str | pathlib.Path, mime_type: str | None = None, handle_404: bool = False, use_last_modified: bool = False):
    mime_type = mime_type or mimetypes.guess_type(path)[0]
    if mime_type is None: raise ValueError("Unknown mime type!")
    ppath = pathlib.Path(path)
    if handle_404 and not ppath.exists():
      return await self.respond_text("not found", 404)

    with open(ppath, "rb") as fd:
      fd_stat = os.stat(fd.fileno())

      if use_last_modified:
        last_modified = formatdate(fd_stat.st_mtime, usegmt=True).encode()
        self.add_response_headers([ (b"Last-Modified", last_modified) ])
        if (last_modified,) == self.headers.get("If-Modified-Since", None):
          await self.response_start(304)
          await self.response_body(b"", False)
          return

      self.add_response_headers(content_headers(fd_stat.st_size, mime_type))
      await self.response_start(200)
      while len(data := fd.read(1_000_000)) != 0:
        await self.response_body(data, fd.tell() != fd_stat.st_size)

  async def receive_json(self): return json.loads(await self.receive_json_raw())

  async def receive_json_raw(self): return await self.receive_text({ "application/json" })

  async def receive_text(self, allowed_mime_types: Iterable[str]):
    allowed_mime_types = allowed_mime_types if isinstance(allowed_mime_types, set) else set(allowed_mime_types)
    mime_type, ct_params = self.content_type
    if mime_type not in allowed_mime_types: raise ValueError(f"Mime type '{mime_type}' is not in allowed types!")
    charset = ct_params.get("charset", "utf-8")
    try: decoder = codecs.getdecoder(charset)
    except LookupError: raise ValueError("Invalid content-type encoding!")
    data = await self.receive_bytes()
    return decoder(data, "ignore")[0]

  async def receive_bytes(self) -> bytes:
    stream = io.BytesIO()
    async for chunk in self.receive_iter():
      _ = stream.write(chunk)
    return stream.getvalue()

  async def receive_iter(self) -> AsyncGenerator[bytes, Any]:
    while True:
      event = await self.receive()
      event_type = event.get("type")
      if event_type == "http.request":
        yield event.get("body", b"")
        if not event.get("more_body", False): return
      elif event_type == "http.disconnect": return

def http_handler(fn: Callable[[HTTPContext], Awaitable[Any]]):
  async def _inner(scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> Any:
    if scope["type"] != "http": raise ASGINextException()
    return await fn(HTTPContext(scope, receive, send))
  return _inner

def websocket_handler(fn: Callable[[WebsocketContext], Awaitable[Any]]):
  async def _inner(scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> Any:
    if scope["type"] != "websocket": raise ASGINextException()
    return await fn(WebsocketContext(scope, receive, send))
  return _inner

CTXT = typing.TypeVar("CTXT", bound=TransportContext)
def routed_handler(pattern: str):
  def _inner(fn: Callable[[CTXT, dict[str, str]], Awaitable[Any]]) -> Callable[[CTXT], Awaitable[Any]]:
    async def _inner_inner(context: CTXT) -> Any:
      if (match:=match_path(pattern, context.path)) is None: context.next()
      return await fn(context, match)
    return _inner_inner
  return _inner

@http_handler
async def http_not_found_handler(context: HTTPContext):
  await context.respond_text("not found", 404)

class Composer:
  def __init__(self) -> None:
    self._handlers: list[ASGIHandler] = []

  def add_handler(self, handler: ASGIHandler):
    self._handlers.append(handler)
    return handler

  async def __call__(self, scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> Any:
    try:
      for handler in self._handlers:
        try: return await handler(scope, receive, send)
        except ASGINextException: pass
    except asyncio.CancelledError: raise
    except BaseException as e:
      logging.debug("asgi error", exc_info=True, stack_info=True)
      if scope["type"] == "websocket":
        return await self._ws_error_handler(WebsocketContext(scope, receive, send), e)
      if scope["type"] == "http":
        return await self._http_error_handler(HTTPContext(scope, receive, send), e)

  async def _http_error_handler(self, context: HTTPContext, exception: BaseException):
    if isinstance(exception, (ValueError, ValidationError)):
      return await context.respond_text("bad request", 400)
    else:
      return await context.respond_text("internal server error", 500)

  async def _ws_error_handler(self, context: WebsocketContext, _exception: BaseException):
    await context.close(1011, "Internal error")



from collections import defaultdict
import unittest, typing
from rxxxt.elements import El, lazy_element
from rxxxt.component import Component, event_handler
from rxxxt.events import ContextInputEvent, CustomOutputEvent, NavigateOutputEvent
from rxxxt.execution import Context
from rxxxt.page import default_page
from rxxxt.session import Session, SessionConfig
from rxxxt.state import JWTStateResolver, local_state, local_state_box

session_config = SessionConfig(page_facotry=default_page, state_resolver=JWTStateResolver(b"deez"), persistent=False)

class TestSession(unittest.IsolatedAsyncioTestCase):
  async def test_state_cell_update(self):
    class Main(Component):
      def render(self):
        return El.div(content=[self.context.path])

    async with Session(session_config, Main()) as session:
      session.set_location("/hello-world")
      await session.init(None)
      update1 = await session.render_update(True, True)
      self.assertIn("/hello-world", update1.html_parts[0])

    async with Session(session_config, Main()) as session:
      await session.init(update1.state_token)
      session.set_location("/world-hello")
      await session.handle_events(()) # this should not matter but we want to match the App flow
      await session.update()
      update2 = await session.render_update(True, True)
      self.assertIn("/world-hello", update2.html_parts[0])

  async def test_cookie_parsing(self):
    @lazy_element
    def main(context: Context):
      return El.div(content=[ context.cookies.get("hello", ""), context.cookies.get("world", "") ])

    async with Session(session_config, main()) as session:
      session.set_location("/")
      session.set_headers({ "cookie": (" hello=world; world=hello",) })
      await session.init(None)

      update = await session.render_update(True, True)
      self.assertIn("worldhello", update.html_parts[0])

  async def test_output_event_refresh(self):
    class Main(Component):
      def render(self):
        return El.div(content=[])

    el = Main()
    async with Session(session_config, el) as session:
      await session.init(None)
      _update1 = await session.render_update(True, True)
      self.assertFalse(session.update_pending)
      el.context.emit("download", { "url": "https://google.com" })
      self.assertTrue(session.update_pending)
      update2 = await session.render_update(True, False)
      self.assertEqual(len(update2.html_parts), 0)
      self.assertEqual(len(update2.events), 1)
      self.assertEqual(update2.events[0], CustomOutputEvent(name="download", data={ "url": "https://google.com" }))

  async def test_partial_update(self):
    class Inner(Component):
      value = local_state(str)

      async def on_init(self) -> None:
        self.value = "1337"

      def render(self): return El.div(content=[self.value])

    class Outer(Component):
      def __init__(self) -> None:
        super().__init__()
        self.inner = Inner()

      def render(self): return El.div(content=[
        "hello",
        self.inner,
        "world"
      ])

    outer = Outer()
    async with Session(session_config, outer) as session:
      session.set_location("/")
      await session.init(None)
      await session.update()
      update1 = await session.render_update(True, True)
      self.assertIn("hello", update1.html_parts[0])
      self.assertIn("world", update1.html_parts[0])
      self.assertIn("1337", update1.html_parts[0])

      outer.inner.value = "133742"
      self.assertIn(outer.inner.context.id, session.state.pending_updates)
      self.assertEqual(len(session.state.pending_updates), 1)
      await session.update()

      update2 = await session.render_update(True, False)
      self.assertEqual(len(update2.html_parts), 1)
      self.assertNotIn("hello", update2.html_parts[0])
      self.assertNotIn("world", update2.html_parts[0])
      self.assertIn("133742", update2.html_parts[0])

    outer = Outer()
    async with Session(session_config, outer) as session:
      session.set_location("/")
      await session.init(update2.state_token)
      outer.inner.value = "247331"
      self.assertIn(outer.inner.context.id, session.state.pending_updates)
      self.assertEqual(len(session.state.pending_updates), 1)
      await session.update()

      update3 = await session.render_update(True, False)
      self.assertEqual(len(update3.html_parts), 1)
      self.assertNotIn("hello", update3.html_parts[0])
      self.assertNotIn("world", update3.html_parts[0])
      self.assertIn("247331", update3.html_parts[0])

  async def test_input_event_handling_order(self):
    event_outputs: list[str] = []
    class Inner(Component):
      @event_handler()
      def on_event(self, value: str): event_outputs.append(value)
      def render(self): return El.div()

    class Outer(Component):
      def __init__(self) -> None:
        super().__init__()
        self.inner = Inner()
      @event_handler()
      def on_event(self, value: str): event_outputs.append(value)
      def render(self): return self.inner

    outer = Outer()
    async with Session(session_config, outer) as session:
      session.set_location("/")
      await session.init(None)
      await session.handle_events((
        ContextInputEvent(context_id=outer.inner.context.sid, data={ "$handler_name": "on_event", "value": "a" }),
        ContextInputEvent(context_id=outer.context.sid, data={ "$handler_name": "on_event", "value": "b" }),
      ))
      await session.update()
      self.assertEqual(event_outputs, ["a", "b"])

  async def test_event_deduplication(self):
    class Main(Component):
      def render(self):
        self.context.navigate("/hello-world")
        self.context.navigate("/world-hello")
        self.context.navigate("/hello-world")
        return El.div()

    async with Session(session_config, Main()) as session:
      session.set_location("/")
      await session.init(None)
      update = await session.render_update(True, True)
      self.assertEqual(update.events, (
        NavigateOutputEvent(location = "/hello-world"),
        NavigateOutputEvent(location = "/world-hello")
      ))

  async def test_deep_state_update(self):
    class Main(Component):
      data = local_state_box(dict[str, typing.Any])
      async def on_init(self) -> None:
        self.data.value = { "hello": "no" }
      def render(self):
        return El.div(content=[ self.data.value.get("hello", "") ])

    el = Main()
    async with Session(session_config, el) as session:
      session.set_location("/")
      await session.init(None)
      el.data.value["hello"] = "yes"
      el.data.update()
      if session.update_pending:
        await session.update()
      update = await session.render_update(True, True)
      self.assertIn("yes", update.html_parts[0])

  async def test_lifecycle(self):
    testobj = self
    counters: defaultdict[str, int] = defaultdict(int)
    class Main(Component):
      async def on_init(self) -> None:
        testobj.assertEqual(counters["before_destroy"], 0)
        counters["init"] += 1
      async def on_before_update(self) -> None:
        testobj.assertEqual(counters["render"], counters["before_update"])
        counters["before_update"] += 1
      def render(self):
        self.context.request_update()
        counters["render"] += 1
        testobj.assertEqual(counters["render"], counters["before_update"])
        return El.div()
      async def on_after_update(self) -> None:
        counters["after_update"] += 1
        testobj.assertEqual(counters["after_update"], counters["render"])
      async def on_before_destroy(self) -> None:
        testobj.assertEqual(counters["init"], 1)
        testobj.assertEqual(counters["before_destroy"], 0)
        counters["before_destroy"] += 1
      async def on_after_destroy(self) -> None:
        counters["after_destroy"] += 1
        testobj.assertEqual(counters["after_destroy"], counters["before_destroy"])

    class Switcher(Component):
      hidden = local_state(bool)
      def render(self):
        if self.hidden: return El.div()
        else: return Main()

    el = Switcher()
    async with Session(session_config, el) as session:
      session.set_location("/")
      await session.init(None)
      await session.update()
      self.assertEqual(counters["render"], 2)
      self.assertEqual(counters["after_destroy"], 0)
      el.hidden = True
      await session.update()
      self.assertEqual(counters["after_destroy"], 1)

if __name__ == "__main__":
  _ = unittest.main()



import unittest

from rxxxt import css_extend

class TestOther(unittest.TestCase):
  def test_css_extend(self):
    self.assertEqual(css_extend({}, class_attr="hello", style_attr="font-size: 10px;"), {
      "class": "hello",
      "style": "font-size: 10px;"
    })

    self.assertEqual(css_extend({ "class": "world", "style": "color:green" }, class_attr="hello", style_attr="font-size: 10px;"), {
      "class": "world hello",
      "style": "color:green;font-size: 10px;"
    })

if __name__ == "__main__":
  _ = unittest.main()



import unittest, httpx, httpx_ws, httpx_ws.transport
from rxxxt.asgi import ASGIHandler, Composer, WebsocketContext, websocket_handler

class TestASGI(unittest.IsolatedAsyncioTestCase):
  def _get_client(self, app: ASGIHandler, websocket: bool = False):
    if websocket: transport = httpx_ws.transport.ASGIWebSocketTransport(app)
    else: transport = httpx.ASGITransport(app=app)
    return httpx.AsyncClient(transport=transport, base_url="http://testserver")

  async def test_ws_error(self):
    composer = Composer()
    @composer.add_handler
    @websocket_handler
    async def _(context: WebsocketContext):
      if context.path != "/hello-world": context.next()
      await context.setup()
      raise RuntimeError()

    @composer.add_handler
    def _(*_): raise RuntimeError()

    client = self._get_client(composer, True)
    with self.assertRaises(httpx_ws.WebSocketDisconnect) as outer_error:
      async with httpx_ws.aconnect_ws(str(client.base_url) + "/world-hello", client): pass

    self.assertEqual(outer_error.exception.code, 1011)

    with self.assertRaises(httpx_ws.WebSocketDisconnect) as inner_error:
      async with httpx_ws.aconnect_ws(str(client.base_url) + "/hello-world", client) as ws:
        _r = await ws.receive_bytes(0.1)

    self.assertEqual(inner_error.exception.code, 1011)

  async def test_ws_messages(self):
    composer = Composer()
    @composer.add_handler
    @websocket_handler
    async def _(context: WebsocketContext):
      await context.setup()
      await context.send_message(b"hello")

    client = self._get_client(composer, True)
    async with httpx_ws.aconnect_ws(str(client.base_url) + "/", client) as ws:
      self.assertEqual(await ws.receive_bytes(), b"hello")
      await ws.close()

  async def test_composer_http_error_400(self):
    composer = Composer()
    @composer.add_handler
    def _(*_): raise ValueError()

    async with self._get_client(composer) as client:
      response = await client.get("/hello-world")
      self.assertEqual(response.status_code, 400)

  async def test_composer_http_error_500(self):
    composer = Composer()
    @composer.add_handler
    def _(*_): raise RuntimeError()

    async with self._get_client(composer) as client:
      response = await client.get("/hello-world")
      self.assertEqual(response.status_code, 500)

if __name__ == "__main__":
  _ = unittest.main()



import asyncio, hashlib, functools, itertools, re
from dataclasses import dataclass
from datetime import datetime
from typing import Literal, Any
from rxxxt.cell import StateCell, StrStateCell
from rxxxt.events import ContextInputEventDescriptor, CustomOutputEvent, EventRegisterQuerySelectorEvent, NavigateOutputEvent, \
  OutputEvent, UseWebsocketOutputEvent, SetCookieOutputEvent, EventRegisterWindowEvent, ContextInputEventDescriptorGenerator
from rxxxt.helpers import T, match_path

ContextStackKey = str | int
ContextStack = tuple[ContextStackKey, ...]

class State:
  """
  State keys may have prefixes. These prefixes inidcate how and when and if a key should be removed from the state.
  Prefixes:
    "#" = temporary - removed from the user data and session state, if no longer used
    "!" = protocol - removed from user state, if not used but not purged from the session
  """

  def __init__(self, update_event: asyncio.Event) -> None:
    self._key_str_store: dict[str, str] = {}
    self._key_cell_store: dict[str, StateCell] = {}

    self._key_subscribers: dict[str, set[ContextStack]] = {}
    self._pending_updates: set[ContextStack] = set()
    self._output_events: list[OutputEvent] = []
    self._update_event = update_event

  @property
  def user_data(self):
    active_keys = self._get_active_keys({ "!", "#" })
    data = { k: v for k, v in self._key_str_store.items() if k in active_keys }
    data.update({ k: cell.serlialize() for k, cell in self._key_cell_store.items() if k in active_keys })
    return data

  @property
  def keys(self) -> set[str]: return set(itertools.chain(self._key_str_store.keys(), self._key_cell_store.keys()))

  @property
  def pending_updates(self):
    return set(self._pending_updates)

  def update(self, k_str_store: dict[str, str]): self._key_str_store.update(k_str_store)

  def get_key_str(self, key: str):
    if (cell := self._key_cell_store.get(key)) is None:
      return self._key_str_store.get(key)
    return cell.serlialize()
  def get_key_cell(self, key: str): return self._key_cell_store.get(key)
  def set_key_cell(self, key: str, cell: StateCell, overwrite: bool = False):
    if key in self._key_cell_store and not overwrite:
      raise ValueError(f"Cell already registered with key '{key}'!")
    self._key_cell_store[key] = cell
  def delete_key(self, key: str):
    _ = self._key_str_store.pop(key, None)
    if (cell := self._key_cell_store.pop(key, None)) is not None:
      cell.destroy()

  def update_state_strs(self, data: dict[str, str]):
    for k, v in data.items():
      cell = self.get_key_cell(k)
      if cell is None:
        self.set_key_cell(k, StrStateCell(v))
      elif isinstance(cell, StrStateCell):
        cell.value = v
      else:
        raise ValueError("invalid cell for location!")
    self.request_key_updates(set(data.keys()))

  def request_context_updates(self, ids: set[ContextStack]):
    for id in ids: self._pending_updates.add(id)
    self._set_update_event()

  def request_key_updates(self, keys: set[str]):
    ids: set[ContextStack] = set(itertools.chain(*(self._key_subscribers.get(key, ()) for key in keys)))
    self.request_context_updates(ids)

  def subscribe(self, cid: ContextStack, key: str):
    self._key_subscribers.setdefault(key, set()).add(cid)

  def unsubscribe(self, cid: ContextStack, key: str):
    subs = self._key_subscribers.get(key)
    if subs is not None and cid in subs: subs.remove(cid)

  def unsubscribe_all(self, cid: ContextStack):
    if cid in self._pending_updates: self._pending_updates.remove(cid)
    for ids in self._key_subscribers.values():
      if cid in ids: ids.remove(cid)
    self._set_update_event()

  def add_output_event(self, event: OutputEvent):
    self._output_events.append(event)
    self._set_update_event()

  def pop_output_events(self):
    res: list[OutputEvent] = []
    for event in self._output_events:
      if event not in res:
        res.append(event)
    self._output_events = []
    return tuple(res)

  def pop_updates(self):
    res = self._pending_updates
    self._pending_updates = set()
    self._set_update_event()
    return res

  def destroy(self):
    for cell in self._key_cell_store.values():
      cell.destroy()
    self._key_cell_store.clear()

  def cleanup(self):
    active_keys = self._get_active_keys({ "#" })
    self._key_str_store = { k: v for k, v in self._key_str_store.items() if k in active_keys }
    inactive_cells = set(self._key_cell_store.keys())
    inactive_cells.difference_update(active_keys)
    for k in inactive_cells:
      if (cell := self._key_cell_store.pop(k, None)) is not None:
        cell.destroy()
    self._key_subscribers = { k: v for k, v in self._key_subscribers.items() if len(v) > 0 }

  def _get_active_keys(self, inactive_prefixes: set[str]):
    active_keys = set(k for k, v in self._key_subscribers.items() if len(v) > 0)
    active_keys.update(k for k in self._key_str_store.keys() if len(k) == 0 or k[0] not in inactive_prefixes)
    active_keys.update(k for k in self._key_cell_store.keys() if len(k) == 0 or k[0] not in inactive_prefixes)
    return active_keys

  def _set_update_event(self):
    if len(self._pending_updates) > 0 or len(self._output_events) > 0: self._update_event.set()
    else: self._update_event.clear()

@functools.lru_cache(maxsize=256)
def get_context_stack_sid(stack: ContextStack):
  hasher = hashlib.sha256()
  for k in stack:
    if isinstance(k, str): k = k.replace(";", ";;")
    else: k = str(k)
    hasher.update((k + ";").encode("utf-8"))
  return hasher.hexdigest()

@dataclass(frozen=True)
class ContextConfig:
  persistent: bool
  render_meta: bool

class Context:
  def __init__(self, state: State, registry: dict[str, Any], config: ContextConfig, stack: ContextStack) -> None:
    self._stack: ContextStack = stack
    self._config = config
    self._registry: dict[str, Any] = dict(registry)
    self.state = state

  @property
  def config(self): return self._config

  @property
  def id(self): return self._stack

  @functools.cached_property
  def sid(self): return get_context_stack_sid(self._stack)

  @property
  def stack_sids(self):
    return [ get_context_stack_sid(self._stack[:i + 1]) for i in range(len(self._stack)) ]

  @property
  def location(self):
    res = self._get_state_str_subscribe("!location")
    if res is None: raise ValueError("No location!")
    else: return res

  @property
  def path(self): return self.location.split("?")[0]

  @property
  def query_string(self):
    parts = self.location.split("?")
    if len(parts) < 2: return None
    else: return parts[1]

  @property
  def cookies(self) -> dict[str, str]:
    values = self.get_header("cookie")
    if len(values) == 0: return {}
    result: dict[str, str] = {}
    for cookie in values[0].split(";"):
      try:
        eq_idx = cookie.index("=")
        result[cookie[:eq_idx].strip()] = cookie[(eq_idx + 1):].strip()
      except ValueError: pass
    return result

  def sub(self, key: ContextStackKey): return Context(self.state, self._registry, self._config, self._stack + (key,))
  def replace_index(self, key: str):
    if isinstance(self._stack[-1], int): return Context(self.state, self._registry, self._config, self._stack[:-1] + (key,))
    if isinstance(self._stack[-1], int): return Context(self.state, self._registry, self._config, self._stack[:-1] + (key,))
    raise ValueError("No index to replace!")
  def update_registry(self, registry: dict[str, Any]): return Context(self.state, self._registry | registry, self._config, self._stack)
  def registered(self, name: str, t: type[T]) -> T:
    if not isinstance((val:=self._registry.get(name)), t):
      raise TypeError(f"Invalid type in get_registered '{type(val)}'!")
    return val

  def match_path(self, pattern: str, re_flags: int = re.IGNORECASE):
    return match_path(pattern, self.path, re_flags)

  def get_header(self, name: str) -> tuple[str, ...]:
    header_lines = self._get_state_str_subscribe(f"!header;{name}")
    if header_lines is None: return ()
    else: return tuple(header_lines.splitlines())

  def request_update(self): self.state.request_context_updates({ self.id })
  def subscribe(self, key: str): self.state.subscribe(self.id, key)
  def unsubscribe(self, key: str): self.state.unsubscribe(self.id, key)
  def unsubscribe_all(self): self.state.unsubscribe_all(self.id)

  def emit(self, name: str, data: dict[str, int | float | str | bool | None]):
    self.state.add_output_event(CustomOutputEvent(name=name, data=data))

  def add_window_event(self, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator):
    self._modify_window_event(name, descriptor, "add")
  def add_query_selector_event(self, selector: str, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator, all: bool = False):
    self._modify_query_selector_event(selector, name, descriptor, all, "add")

  def remove_window_event(self, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator):
    self._modify_window_event(name, descriptor, "remove")
  def remove_query_selector_event(self, selector: str, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator, all: bool = False):
    self._modify_query_selector_event(selector, name, descriptor, all, "remove")

  def navigate(self, location: str):
    is_full_url = ":" in location # colon means full url
    if not is_full_url: self.state.update_state_strs({"!location": location})
    self.state.add_output_event(NavigateOutputEvent(location=location, requires_refresh=is_full_url))
  def use_websocket(self, websocket: bool = True): self.state.add_output_event(UseWebsocketOutputEvent(websocket=websocket))
  def set_cookie(self, name: str, value: str, expires: datetime | None = None, path: str | None = None,
                secure: bool | None = None, http_only: bool | None = None, domain: str | None = None, max_age: int | None = None, mirror_state: bool = True):
    self.state.add_output_event(SetCookieOutputEvent(name=name, value=value, expires=expires, path=path, secure=secure, http_only=http_only, domain=domain, max_age=max_age))
    if mirror_state:
      self.state.update_state_strs({ "!header;cookie": "; ".join(f"{k}={v}" for k, v in (self.cookies | { name: value }).items()) })
  def delete_cookie(self, name: str, mirror_state: bool = True):
    self.state.add_output_event(SetCookieOutputEvent(name=name, max_age=-1))
    if mirror_state:
      self.state.update_state_strs({ "!header;cookie": "; ".join(f"{k}={v}" for k, v in self.cookies.items() if k != name) })
  def _get_state_str_subscribe(self, key: str):
    res = self.state.get_key_str(key)
    self.state.subscribe(self.id, key)
    return res
  def _modify_window_event(self, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator, mode: Literal["add"] | Literal["remove"]):
    descriptor = descriptor.descriptor if isinstance(descriptor, ContextInputEventDescriptorGenerator) else descriptor
    self.state.add_output_event(EventRegisterWindowEvent(name=name, mode=mode, descriptor=descriptor))
  def _modify_query_selector_event(self, selector: str, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator, all: bool, mode: Literal["add"] | Literal["remove"]):
    descriptor = descriptor.descriptor if isinstance(descriptor, ContextInputEventDescriptorGenerator) else descriptor
    self.state.add_output_event(EventRegisterQuerySelectorEvent(
      name=name,
      mode=mode,
      selector=selector,
      all=all,
      descriptor=descriptor))



from abc import ABC, abstractmethod
import html
import logging
from typing import Callable, Concatenate, Generic, Protocol
from collections.abc import Iterable
from rxxxt.execution import Context
from rxxxt.helpers import FNP
from rxxxt.node import ElementNode, FragmentNode, Node, TextNode, VoidElementNode
from typing import Any

class Element(ABC):
  @abstractmethod
  def tonode(self, context: Context) -> Node: ...

class CustomAttribute(ABC):
  @abstractmethod
  def get_key_values(self, original_key: str) -> tuple[tuple[str, str | None],...]: ...

ElementContent = Iterable[Element | str]
HTMLAttributeValue = str | bool | int | float | CustomAttribute | None
HTMLAttributes = dict[str, str | bool | int | float | CustomAttribute | None]

def _elements_to_ordered_nodes(context: Context, elements: tuple[Element, ...]):
  return tuple(el.tonode(context.sub(idx)) for idx, el in enumerate(elements))

def _element_content_to_elements(content: ElementContent) -> tuple[Element, ...]:
  return tuple(TextElement(item) if isinstance(item, str) else item for item in content)

def _normalize_attrs(attrs: HTMLAttributes):
  return { k.lstrip("_"): v for k,v in attrs.items() }

def _apply_attribute_kv(attributes: dict[str, str | None], k: str, v: Any):
  if isinstance(v, bool):
    if not v: return
    v = None
  elif isinstance(v, (int, float)): v = str(v)
  if v is not None and not isinstance(v, str):
    raise ValueError("Invalid attribute value", v)
  attributes[k] = v

class HTMLFragment(Element):
  def __init__(self, content: ElementContent) -> None:
    super().__init__()
    self._content = _element_content_to_elements(content)

  def tonode(self, context: Context) -> Node:
    return FragmentNode(context, _elements_to_ordered_nodes(context, self._content))

class HTMLVoidElement(Element):
  def __init__(self, tag: str, attributes: HTMLAttributes) -> None:
    super().__init__()
    self._tag = tag
    self._attributes: dict[str, str | None] = {}
    for k, v in attributes.items():
      if isinstance(v, CustomAttribute):
        for ik, iv in v.get_key_values(k):
          _apply_attribute_kv(self._attributes, ik, iv)
      else:
        _apply_attribute_kv(self._attributes, k, v)

  def tonode(self, context: Context) -> 'Node':
    return VoidElementNode(context, self._tag, self._attributes)

class HTMLElement(HTMLVoidElement):
  def __init__(self, tag: str, attributes: HTMLAttributes, content: ElementContent) -> None:
    super().__init__(tag, attributes)
    self._content = _element_content_to_elements(content)

  def tonode(self, context: Context) -> 'Node':
    return ElementNode(context, self._tag, self._attributes, _elements_to_ordered_nodes(context, self._content))

class KeyedElement(Element):
  def __init__(self, key: str, element: Element) -> None:
    super().__init__()
    self._key = key
    self._element = element

  def tonode(self, context: Context) -> 'Node':
    try: context = context.replace_index(self._key)
    except ValueError as e: logging.debug(f"Failed to replace index with key {self._key}", e)
    return self._element.tonode(context)

class TaggedElement(Element):
  def __init__(self, tag: str, element: Element) -> None:
    super().__init__()
    self._tag = tag
    self._element = element

  def tonode(self, context: Context) -> 'Node':
    return self._element.tonode(context = context.sub(self._tag))

class WithRegistered(Element):
  def __init__(self, register: dict[str, Any], child: Element) -> None:
    super().__init__()
    self._register = register
    self._child = child

  def tonode(self, context: Context) -> 'Node':
    return self._child.tonode(context.update_registry(self._register))

class _LazyElement(Element, Generic[FNP]):
  def __init__(self, fn: Callable[Concatenate[Context, FNP], Node], *args: FNP.args, **kwargs: FNP.kwargs) -> None:
    self._fn = fn
    self._fn_args = args
    self._fn_kwargs = kwargs

  def tonode(self, context: Context) -> 'Node':
    return self._fn(context, *self._fn_args, **self._fn_kwargs)

def lazy_node(fn: Callable[Concatenate[Context, FNP], Node]) -> Callable[FNP, 'Element']:
  def _inner(*args: FNP.args, **kwargs: FNP.kwargs) -> Element:
    return _LazyElement(fn, *args, **kwargs)
  return _inner

def lazy_element(fn: Callable[Concatenate[Context, FNP], Element]) -> Callable[FNP, 'Element']:
  def _inner(context: Context, *args: FNP.args, **kwargs: FNP.kwargs) -> Node:
    return fn(context, *args, **kwargs).tonode(context)
  return lazy_node(_inner)

@lazy_node
def TextElement(context: Context, text: str):
  return TextNode(context, html.escape(text))

@lazy_node
def UnescapedHTMLElement(context: Context, text: str):
  return TextNode(context, text)

@lazy_node
def ScriptContent(context: Context, script: str):
  return TextNode(context, script.replace("</", "<\\/"))

class CreateHTMLElement(Protocol):
  def __call__(self, content: ElementContent = (), key: str | None = None, **kwargs: HTMLAttributeValue) -> Element: ...

class _El(type):
  def __getitem__(cls, name: str) -> CreateHTMLElement:
    def _inner(content: ElementContent = (), key: str | None = None, **kwargs: HTMLAttributeValue):
      el = HTMLElement(name, attributes=_normalize_attrs(kwargs), content=list(content))
      if key is not None: el = KeyedElement(key, el)
      return el
    return _inner
  def __getattribute__(cls, name: str):
    return cls[name]

class El(metaclass=_El): ...

class CreateHTMLVoidElement(Protocol):
  def __call__(self, key: str | None = None, **kwargs: HTMLAttributeValue) -> Element: ...

class _VEl(type):
  def __getitem__(cls, name: str) -> CreateHTMLVoidElement:
    def _inner(key: str | None = None, **kwargs: HTMLAttributeValue) -> Element:
      el = HTMLVoidElement(name, attributes=_normalize_attrs(kwargs))
      if key is not None: el = KeyedElement(key, el)
      return el
    return _inner
  def __getattribute__(cls, name: str):
    return cls[name]

class VEl(metaclass=_VEl): ...

class ElementFactory(Protocol):
  def __call__(self) -> Element: ...

def meta_element(id: str, inner: Element):
  return HTMLElement("rxxxt-meta", {"id":id}, [inner])

def class_map(map: dict[str, bool]):
  return " ".join([ k for k, v in map.items() if v ])

def css_extend(attrs: HTMLAttributes, class_attr: str = "", style_attr: str = ""):
  attrs = _normalize_attrs(attrs)
  if class_attr:
    if "class" in attrs: attrs["class"] = (str(attrs["class"]) + " " + class_attr).strip()
    else: attrs["class"] = class_attr
  if style_attr:
    if "style" in attrs: attrs["style"] = (str(attrs["style"]) + ";" + style_attr).strip()
    else: attrs["style"] = style_attr
  return attrs



import asyncio, contextlib, importlib.resources
from typing import Any, Literal
from pydantic import BaseModel
from rxxxt.asgi import ASGIFnReceive, ASGIFnSend, ASGIScope, Composer, HTTPContext, WebsocketContext, http_handler, http_not_found_handler, routed_handler, websocket_handler
from rxxxt.elements import ElementFactory
from rxxxt.events import InputEvent
from rxxxt.page import PageFactory, default_page
from rxxxt.session import Session, SessionConfig
from rxxxt.state import StateResolver, default_state_resolver

class AppHttpRequest(BaseModel):
  state_token: str
  events: tuple[InputEvent, ...]

class AppWebsocketInitMessage(BaseModel):
  type: Literal["init"]
  state_token: str
  enable_state_updates: bool

class AppWebsocketUpdateMessage(BaseModel):
  type: Literal["update"]
  events: tuple[InputEvent, ...]
  location: str

class App:
  def __init__(self, content: ElementFactory, state_resolver: StateResolver | None = None, page_factory: PageFactory = default_page) -> None:
    self._content = content
    self._page_factory: PageFactory = page_factory
    self._state_resolver = state_resolver or default_state_resolver()
    self._composer = Composer()
    _ = self._composer.add_handler(http_handler(routed_handler("/rxxxt-client.js")(self._http_static_rxxxt_client_js)))
    _ = self._composer.add_handler(http_handler(self._http_session))
    _ = self._composer.add_handler(websocket_handler(self._ws_session))
    _ = self._composer.add_handler(http_not_found_handler)

  async def __call__(self, scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> Any:
    return await self._composer(scope, receive, send)

  async def _ws_session(self, context: WebsocketContext):
    await context.setup()
    message = await context.receive_message()
    init_message = AppWebsocketInitMessage.model_validate_json(message)

    with contextlib.suppress(ConnectionError):
      async with Session(self._get_session_config(True), self._content()) as session:
        updating_lock = asyncio.Lock()

        async def updater():
          while context.connected:
            await session.wait_for_update()
            async with updating_lock:
              await session.update()
              data = await session.render_update(include_state_token=init_message.enable_state_updates, render_full=False)
              await context.send_message(data.model_dump_json(exclude_defaults=True))

        await session.init(init_message.state_token)

        session.set_location(context.location)
        session.set_headers(context.headers)

        updater_task = asyncio.create_task(updater())
        try:
          while True:
            message = await context.receive_message()
            async with updating_lock:
              update_message = AppWebsocketUpdateMessage.model_validate_json(message)
              session.set_location(update_message.location)
              await session.handle_events(update_message.events)
        finally: _ = updater_task.cancel()

  async def _http_session(self, context: HTTPContext):
    if context.method not in ("POST", "GET"): context.next()
    async with Session(self._get_session_config(False), self._content()) as session:
      if context.method == "POST":
        req = AppHttpRequest.model_validate_json(await context.receive_json_raw())
        await session.init(req.state_token)
        session.set_location(context.location)
        session.set_headers(context.headers)
        await session.handle_events(req.events)
      else:
        session.set_location(context.location)
        session.set_headers(context.headers)
        await session.init(None)

      if session.update_pending:
        await session.update()

      if context.method == "POST":
        result = await session.render_update(include_state_token=True, render_full=False)
        await context.respond_text(result.model_dump_json(exclude_defaults=True), mime_type="application/json")
      else:
        result = await session.render_page(context.path)
        await context.respond_text(result, mime_type="text/html")

  async def _http_static_rxxxt_client_js(self, context: HTTPContext, _: dict[str, str]):
    with importlib.resources.path("rxxxt.assets", "main.js") as file_path:
      return await context.respond_file(file_path, use_last_modified=True)

  def _get_session_config(self, persistent: bool):
    return SessionConfig(page_facotry=self._page_factory, state_resolver=self._state_resolver, persistent=persistent)
